import sys
import os
import pytest

# from pytest_lazyfixture import lazy_fixture as lf


thisfile = os.path.abspath(__file__)
modulepath = os.path.dirname(os.path.dirname(thisfile))

sys.path.insert(0, modulepath)
import tensorcircuit as tc


def test_qsikit_compiler():
    try:
        import qiskit as _
    except ImportError:
        pytest.skip("qiskit is not installed")

    from tensorcircuit.compiler.qiskit_compiler import qiskit_compile

    c = tc.Circuit(2)
    c.x(1)
    c.cx(0, 1)

    c1, info = qiskit_compile(
        c,
        info=None,
        output="qasm",
        compiled_options={
            "basis_gates": ["cz", "rz", "h"],
            "optimization_level": 3,
            "coupling_map": [[0, 2], [2, 0], [1, 0], [0, 1]],
        },
    )
    assert "cz" in c1
    print(info["logical_physical_mapping"])

    c = tc.Circuit(2)
    c.x(1)
    c.cx(0, 1)
    c.measure_instruction(1)
    c1, info = qiskit_compile(
        c,
        info=None,
        output="tc",
        compiled_options={
            "basis_gates": ["cx", "rz", "h"],
            "optimization_level": 3,
            "coupling_map": [[0, 2], [2, 0], [1, 0], [0, 1]],
            "initial_layout": [1, 2],
        },
    )
    for inst in c1.to_qir():
        if inst["name"] == "h":
            assert inst["index"][0] == 2
    print(c1.draw())
    assert info["logical_physical_mapping"][1] in [0, 2]
    print(info)
    c2, info2 = qiskit_compile(
        c1,
        info=info,
        output="tc",
        compiled_options={
            "basis_gates": ["cx", "rz", "h"],
            "optimization_level": 3,
            "coupling_map": [[0, 2], [2, 0], [1, 0], [0, 1]],
        },
    )
    assert info2["positional_logical_mapping"] == {0: 1}
    print(c2.draw())


def test_composed_compiler():
    from tensorcircuit.compiler import default_compiler

    c = tc.Circuit(3)
    c.rx(0)
    c.cx(0, 1)
    c.cz(1, 0)
    c.rxx(0, 2, theta=0.2)
    c.measure_instruction(2)
    c.measure_instruction(0)
    c1, info = default_compiler(c)
    print(c1.draw())
    assert c1.gate_count_by_condition(lambda qir: qir["name"] == "cnot") == 4
    assert info["positional_logical_mapping"][0] == 2

    default_compiler.add_options(
        {
            "basis_gates": ["h", "rz", "cz"],
            "optimization_level": 2,
            "coupling_map": [[0, 1], [1, 2]],
        }
    )
    c1, info = default_compiler(c)
    assert c1.gate_count_by_condition(lambda qir: qir["name"] == "cnot") == 0
    print(info)



"""
Abstraction for Provider, Device and Task
"""

from typing import Any, Dict, List, Optional, Union, Tuple
from functools import partial
import time

import networkx as nx

from ..results import readout_mitigation as rem
from ..results import counts
from ..utils import arg_alias


class TCException(BaseException):
    pass


class TaskException(TCException):
    pass


class TaskUnfinished(TaskException):
    def __init__(self, taskid: str, state: str):
        self.taskid = taskid
        self.state = state
        super().__init__(
            "Task %s is not completed yet, now in %s state" % (self.taskid, self.state)
        )


class TaskFailed(TaskException):
    def __init__(self, taskid: str, state: str, message: str):
        self.taskid = taskid
        self.state = state
        self.message = message
        super().__init__(
            "Task %s is in %s state with err message %s"
            % (self.taskid, self.state, self.message)
        )


class Provider:
    """
    Provider abstraction for cloud connection, eg. "tencent", "local"
    """

    activated_providers: Dict[str, "Provider"] = {}

    def __init__(self, name: str, lower: bool = True):
        if lower is True:
            name = name.lower()
        self.name = name

    def __str__(self) -> str:
        return self.name

    __repr__ = __str__

    @classmethod
    def from_name(cls, provider: Union[str, "Provider"] = "tencent") -> "Provider":
        if provider is None:
            provider = "tencent"
        if isinstance(provider, cls):
            p = provider
        elif isinstance(provider, str):
            if provider in cls.activated_providers:
                return cls.activated_providers[provider]
            else:
                p = cls(provider)
                cls.activated_providers[provider] = p
        else:
            raise ValueError(
                "Unsupported format for `provider` argument: %s" % provider
            )
        return p

    def set_token(self, token: str, cached: bool = True) -> Any:
        from .apis import set_token

        return set_token(token, self, cached=cached)

    def get_token(self) -> str:
        from .apis import get_token

        return get_token(self)  # type: ignore

    def list_devices(self, **kws: Any) -> Any:
        from .apis import list_devices

        return list_devices(self, **kws)

    def get_device(self, device: Optional[Union[str, "Device"]]) -> "Device":
        from .apis import get_device

        return get_device(self, device)

    def list_tasks(self, **filter_kws: Any) -> List["Task"]:
        from .apis import list_tasks

        return list_tasks(self, **filter_kws)


sep = "::"


class Device:
    """
    Device abstraction for cloud connection, eg. quantum chips
    """

    activated_devices: Dict[str, "Device"] = {}

    def __init__(
        self,
        name: str,
        provider: Optional[Union[str, Provider]] = None,
        lower: bool = True,
    ):
        if lower is True:
            name = name.lower()
        if provider is not None:
            self.provider = Provider.from_name(provider)
            if len(name.split(sep)) > 1:
                self.name = name.split(sep)[1]
            else:
                self.name = name
        else:  # no explicit provider
            if len(name.split(sep)) > 1:
                self.name = name.split(sep)[1]
                self.provider = Provider.from_name(name.split(sep)[0])
            else:
                from .apis import get_provider

                self.name = name
                self.provider = get_provider()
        self.readout_mit: Any = None

    def __str__(self) -> str:
        return self.provider.name + sep + self.name

    __repr__ = __str__

    @classmethod
    def from_name(
        cls,
        device: Union[str, "Device"],
        provider: Optional[Union[str, Provider]] = None,
    ) -> "Device":
        # if device is None:
        # raise ValueError("Must specify on device instead of default ``None``")
        if isinstance(device, cls):
            d = device
        elif isinstance(device, str):
            if len(device.split(sep)) > 1:
                provider = device.split(sep)[0]
                device = device.split(sep)[1]
            if provider is None:
                pn = ""
            elif isinstance(provider, str):
                pn = provider
            else:
                pn = provider.name
            if pn + sep + device in cls.activated_devices:
                return cls.activated_devices[pn + sep + device]
            else:
                d = cls(device, provider)

                cls.activated_devices[pn + sep + device] = d
        else:
            raise ValueError("Unsupported format for `provider` argument: %s" % device)

        return d

    def set_token(self, token: str, cached: bool = True) -> Any:
        from .apis import set_token

        return set_token(token, provider=self.provider, device=self, cached=cached)

    def get_token(self) -> Optional[str]:
        from .apis import get_token

        s = get_token(provider=self.provider, device=self)
        if s is not None:
            return s
        # fallback to provider default
        return get_token(provider=self.provider)

    def list_properties(self) -> Dict[str, Any]:
        """
        List all device properties in as dict

        :return: [description]
        :rtype: Dict[str, Any]
        """
        from .apis import list_properties

        return list_properties(self.provider, self)

    def native_gates(self) -> List[str]:
        """
        List native gates supported for the device, str conforms qiskit convention

        :return: _description_
        :rtype: List[str]
        """
        properties = self.list_properties()

        if "native_gates" in properties:
            return properties["native_gates"]  # type: ignore
        return []

    def topology(self) -> List[Tuple[int, int]]:
        """
        Get the bidirectional topology link list of the device

        :return: [description]
        :rtype: List[Tuple[int, int]]
        """
        properties = self.list_properties()
        if "links" not in properties:
            return  # type: ignore
        links = []
        for link in properties["links"]:
            links.append((link[0], link[1]))
            links.append((link[1], link[0]))
        links = list(set(links))
        links = [list(link) for link in links]  # type: ignore
        # compatible with coupling_map in qiskit
        return links

    def topology_graph(self, visualize: bool = False) -> nx.Graph:
        """
        Get the qubit topology in ``nx.Graph`` or directly visualize it

        :param visualize: [description], defaults to False
        :type visualize: bool, optional
        :return: [description]
        :rtype: nx.Graph
        """
        pro = self.list_properties()
        if not ("links" in pro and "bits" in pro):
            return  # type: ignore
        g = nx.Graph()
        node_color = []
        edge_color = []
        for i in pro["bits"]:
            g.add_node(i)
            node_color.append(pro["bits"][i]["T1"])
        for e1, e2 in pro["links"]:
            g.add_edge(e1, e2)
            edge_color.append(pro["links"][(e1, e2)]["CZErrRate"])
        if visualize is False:
            return g
        from matplotlib import colormaps

        # pos1 = nx.planar_layout(g)
        # pos2 = nx.spring_layout(g, pos=pos1, k=2)
        pos = nx.kamada_kawai_layout(g)
        return nx.draw(
            g,
            pos=pos,
            with_labels=True,
            node_size=600,
            node_color=node_color,
            cmap=colormaps["Wistia"],
            vmin=max(min(node_color) - 5, 0),
            width=2.5,
            edge_color=edge_color,
            edge_cmap=colormaps["gray"],
            edge_vmin=0,
            edge_vmax=max(edge_color) * 1.2,
        )

    def get_task(self, taskid: str) -> "Task":
        from .apis import get_task

        return get_task(taskid, device=self)

    def submit_task(self, **task_kws: Any) -> List["Task"]:
        from .apis import submit_task

        return submit_task(provider=self.provider, device=self, **task_kws)

    def list_tasks(self, **filter_kws: Any) -> List["Task"]:
        from .apis import list_tasks

        return list_tasks(self.provider, self, **filter_kws)


sep2 = "~~"


class Task:
    """
    Task abstraction for quantum jobs on the cloud
    """

    def __init__(self, id_: str, device: Optional[Device] = None):
        self.id_ = id_
        self.device = device
        self.more_details: Dict[str, Any] = {}

    def __repr__(self) -> str:
        return self.device.__repr__() + sep2 + self.id_

    __str__ = __repr__

    def get_device(self) -> Device:
        """
        Query which device the task is run on

        :return: _description_
        :rtype: Device
        """
        if self.device is None:
            return Device.from_name(self.details()["device"])
        else:
            return Device.from_name(self.device)

    @partial(arg_alias, alias_dict={"blocked": ["wait"]})
    def details(self, blocked: bool = False, **kws: Any) -> Dict[str, Any]:
        """
        Get the current task details


        :param blocked: whether return until task is finished, defaults to False
        :type blocked: bool
        :return: _description_
        :rtype: Dict[str, Any]
        """
        from .apis import get_task_details

        if blocked is False:
            dt = get_task_details(self, **kws)
            dt.update(self.more_details)
            return dt
        s = self.state()
        tries = 0
        while s == "pending":
            time.sleep(0.5 + tries / 10)
            tries += 1
            s = self.state()
        return self.details(**kws)  # type: ignore

    def get_logical_physical_mapping(self) -> Optional[Dict[int, int]]:
        d = self.details()
        try:
            mp = d["optimization"]["pairs"]
        except KeyError:
            if "qubits" in d and isinstance(d["qubits"], int):
                mp = {i: i for i in range(d["qubits"])}
            else:
                mp = None
        return mp  # type: ignore

    def add_details(self, **kws: Any) -> None:
        self.more_details.update(kws)

    def state(self) -> str:
        """
        Query the current task status

        :return: _description_
        :rtype: str
        """
        r = self.details()
        return r["state"]  # type: ignore

    status = state

    def resubmit(self) -> "Task":
        """
        resubmit the task

        :return: the resubmitted task
        :rtype: Task
        """
        from .apis import resubmit_task

        return resubmit_task(self)

    @partial(arg_alias, alias_dict={"format": ["format_"], "blocked": ["wait"]})
    def results(
        self,
        format: Optional[str] = None,
        blocked: bool = True,
        mitigated: bool = False,
        calibriation_options: Optional[Dict[str, Any]] = None,
        readout_mit: Optional[rem.ReadoutMit] = None,
        mitigation_options: Optional[Dict[str, Any]] = None,
    ) -> counts.ct:
        """
        get task results of the qjob

        :param format: unsupported now, defaults to None, which is "count_dict_bin"
        :type format: Optional[str], optional
        :param blocked: whether blocked to wait until the result is returned, defaults to False,
            which raise error when the task is unfinished
        :type blocked: bool, optional
        :param mitigated: whether enable readout error mitigation, defaults to False
        :type mitigated: bool, optional
        :param calibriation_options: option dict for ``ReadoutMit.cals_from_system``,
            defaults to None
        :type calibriation_options: Optional[Dict[str, Any]], optional
        :param readout_mit: if given, directly use the calibriation info on ``readout_mit``,
            defaults to None
        :type readout_mit: Optional[rem.ReadoutMit], optional
        :param mitigation_options: option dict for ``ReadoutMit.apply_correction``, defaults to None
        :type mitigation_options: Optional[Dict[str, Any]], optional
        :return: count dict results
        :rtype: Any
        """
        if not blocked:
            s = self.state()
            if s != "completed":
                raise TaskUnfinished(self.id_, s)
            r = self.details()["results"]
            r = counts.sort_count(r)  # type: ignore
        else:
            s = self.state()
            tries = 0
            while s != "completed":
                if s in ["failed"]:
                    err = self.details().get("err", "")
                    raise TaskFailed(self.id_, s, err)
                time.sleep(0.5 + tries / 10)
                tries += 1
                s = self.state()
            r = self.results(format=format, blocked=False, mitigated=False)
        if mitigated is False:
            return r  # type: ignore
        nqubit = len(list(r.keys())[0])

        # mitigated is True:
        device = self.get_device()
        if device.provider.name != "tencent":
            raise ValueError("Only tencent provider supports auto readout mitigation")
        if readout_mit is None and getattr(device, "readout_mit", None) is None:

            def run(cs: Any, shots: Any) -> Any:
                """
                current workaround for batch
                """
                from .apis import submit_task

                ts = submit_task(circuit=cs, shots=shots, device=device.name + "?o=0")
                return [t.results(blocked=True) for t in ts]  # type: ignore

            shots = self.details()["shots"]
            readout_mit = rem.ReadoutMit(run)
            if calibriation_options is None:
                calibriation_options = {}
            readout_mit.cals_from_system(
                list(range(nqubit)), shots, **calibriation_options
            )
            device.readout_mit = readout_mit
        elif readout_mit is None:
            readout_mit = device.readout_mit

        if mitigation_options is None:
            try:
                mitigation_options = {
                    "logical_physical_mapping": self.details()["optimization"]["pairs"]
                }
            except KeyError:
                mitigation_options = {}
        miti_count = readout_mit.apply_correction(
            r, list(range(nqubit)), **mitigation_options
        )
        return counts.sort_count(miti_count)



"""
higher level wrapper shortcut for submit_task
"""
from typing import Any, Callable, Dict, List, Optional, Sequence, Union
import logging
import time

import numpy as np

from ..circuit import Circuit
from ..results import counts
from ..results.readout_mitigation import ReadoutMit
from ..utils import is_sequence
from ..cons import backend
from ..quantum import ps2xyz
from ..compiler.qiskit_compiler import qiskit_compile
from .apis import submit_task, get_device
from .abstraction import Device


logger = logging.getLogger(__name__)
Tensor = Any


def batch_submit_template(device: str) -> Callable[..., List[counts.ct]]:
    # TODO(@refraction-ray): fixed when batch submission really works
    def run(cs: Union[Circuit, Sequence[Circuit]], shots: int) -> List[counts.ct]:
        """
        batch circuit running alternative
        """
        single = False
        if not is_sequence(cs):
            cs = [cs]  # type: ignore
            single = True
        ts = []
        # for c in cs:  # type: ignore
        #     ts.append(submit_task(circuit=c, shots=shots, device=device))
        #     time.sleep(0.3)
        ts = submit_task(circuit=cs, shots=shots, device=device)
        l = [t.results(blocked=True) for t in ts]  # type: ignore
        if single is False:
            return l
        return l[0]  # type: ignore

    return run


def sample_expectation_ps(
    c: Circuit,
    x: Optional[Sequence[int]] = None,
    y: Optional[Sequence[int]] = None,
    z: Optional[Sequence[int]] = None,
    shots: int = 1024,
    device: Optional[Device] = None,
    **kws: Any,
) -> float:
    # deprecated
    # TODO(@refraction-ray): integrated error mitigation
    c1 = Circuit.from_qir(c.to_qir())
    if x is None:
        x = []
    if y is None:
        y = []
    if z is None:
        z = []
    for i in x:
        c1.H(i)  # type: ignore
    for i in y:
        c1.rx(i, theta=np.pi / 2)  # type: ignore
    if device is None:
        device = get_device()
    t = submit_task(circuit=c1, device=device, shots=shots)
    raw_counts = t.results(blocked=True)  # type: ignore
    x, y, z = list(x), list(y), list(z)
    return counts.expectation(raw_counts, x + y + z)


def batch_expectation_ps(
    c: Circuit,
    pss: List[List[int]],
    device: Any = None,
    ws: Optional[List[float]] = None,
    shots: int = 8192,
    with_rem: bool = True,
) -> Union[Any, List[Any]]:
    if device is None:
        results = []
        for ps in pss:
            results.append(c.expectation_ps(**ps2xyz(ps)))  # type: ignore
        if ws is None:
            return backend.stack(results)
        else:
            return backend.sum([w * r for w, r in zip(ws, results)])
    cs = []
    infos = []
    exps = []
    if isinstance(device, str):
        device = get_device(device)
    for ps in pss:
        # TODO(@refraction-ray): Pauli string grouping
        c1 = Circuit.from_qir(c.to_qir())
        exp = []
        for j, i in enumerate(ps):
            if i == 1:
                c1.H(j)  # type: ignore
                exp.append(j)
            elif i == 2:
                c1.rx(j, theta=np.pi / 2)  # type: ignore
                exp.append(j)
            elif i == 3:
                exp.append(j)
        c1, info = qiskit_compile(
            c1,
            compiled_options={
                "basis_gates": device.native_gates(),
                "optimization_level": 3,
                "coupling_map": device.topology(),
            },
        )
        cs.append(c1)
        infos.append(info)
        exps.append(exp)

    reduced_cs = []
    reduced_dict = {}
    recover_dict = {}
    # merge the same circuit
    for j, ps in enumerate(pss):
        ps = [i if i in [1, 2] else 0 for i in ps]
        if tuple(ps) not in reduced_dict:
            reduced_dict[tuple(ps)] = [j]
            reduced_cs.append(cs[j])
            recover_dict[tuple(ps)] = len(reduced_cs) - 1
        else:
            reduced_dict[tuple(ps)].append(j)

    def run(cs: List[Any], shots: int) -> List[Dict[str, int]]:
        logger.info(f"submit task on {device.name} for {len(cs)} circuits")
        time0 = time.time()
        ts = submit_task(
            circuit=cs,
            device=device,
            shots=shots,
            enable_qos_qubit_mapping=False,
            enable_qos_gate_decomposition=False,
        )
        if not is_sequence(ts):
            ts = [ts]  # type: ignore
        raw_counts = [t.results(blocked=True) for t in ts]
        time1 = time.time()
        logger.info(
            f"finished collecting count results of {len(cs)} tasks in {round(time1-time0, 4)} seconds"
        )
        return raw_counts

    reduced_raw_counts = run(reduced_cs, shots)
    raw_counts: List[Dict[str, int]] = [None] * len(cs)  # type: ignore
    for i in range(len(cs)):
        ps = [i if i in [1, 2] else 0 for i in pss[i]]
        raw_counts[i] = reduced_raw_counts[recover_dict[tuple(ps)]]

    if with_rem:
        if getattr(device, "readout_mit", None) is None:
            mit = ReadoutMit(run)
            # TODO(@refraction-ray) only work for tencent provider
            nq = device.list_properties().get("qubits", None)
            if nq is None:
                nq = c._nqubits
            mit.cals_from_system(nq, shots=shots)
            device.readout_mit = mit
        else:
            mit = device.readout_mit

        results = [
            mit.expectation(raw_counts[i], exps[i], **infos[i])
            for i in range(len(raw_counts))
        ]
    else:
        results = [
            counts.expectation(raw_counts[i], exps[i]) for i in range(len(raw_counts))
        ]
    if ws is not None:
        sumr = sum([w * r for w, r in zip(ws, results)])
        return sumr
    return results



"""
Customized ops for ML framework
"""
# pylint: disable=invalid-name

from typing import Any

import torch  # type: ignore

Array = Any  # torch Tensor

qr_epsilon = 1e-8


def torchqr_grad(
    a: Array,
    q: Array,
    r: Array,
    dq: Array,
    dr: Array,
) -> Array:
    """Get the gradient for Qr."""

    if r.shape[-2] > r.shape[-1] and q.shape[-2] == q.shape[-1]:
        raise NotImplementedError(
            "QrGrad not implemented when nrows > ncols "
            "and full_matrices is true. Received r.shape="
            f"{r.shape} with nrows={r.shape[-2]}"
            f"and ncols={r.shape[-1]}."
        )

    def _TriangularSolve(x: Array, r: Array) -> Array:
        """Equivalent to matmul(x, adjoint(matrix_inverse(r))) if r is upper-tri."""
        return torch.linalg.solve_triangular(
            r, x.adjoint(), upper=True, unitriangular=False
        ).adjoint()

    def _QrGradSquareAndDeepMatrices(q: Array, r: Array, dq: Array, dr: Array) -> Array:
        """
        Get the gradient for matrix orders num_rows >= num_cols and full_matrices is false.
        """

        # Modification begins
        rdiag = torch.linalg.diagonal(r)
        # if abs(rdiag[i]) < qr_epsilon then rdiag[i] = qr_epsilon otherwise keep the old value
        qr_epsilon_diag = torch.ones_like(rdiag) * qr_epsilon
        rdiag = torch.where(rdiag.abs() < qr_epsilon, qr_epsilon_diag, rdiag)
        r = torch.diagonal_scatter(r, rdiag, dim1=-2, dim2=-1)
        # delta_dq = math_ops.matmul(q, math_ops.matmul(dr, tf.linalg.adjoint(delta_r)))
        # dq = dq + delta_dq
        # Modification ends

        qdq = torch.matmul(q.adjoint(), dq)
        qdq_ = qdq - qdq.adjoint()
        rdr = torch.matmul(r, dr.adjoint())
        rdr_ = rdr - rdr.adjoint()
        tril = torch.tril(qdq_ + rdr_)

        grad_a = torch.matmul(q, dr + _TriangularSolve(tril, r))
        grad_b = _TriangularSolve(dq - torch.matmul(q, qdq), r)
        ret = grad_a + grad_b

        if q.is_complex():
            m = rdr - qdq.adjoint()
            eyem = torch.diagonal_scatter(
                torch.zeros_like(m), torch.linalg.diagonal(m), dim1=-2, dim2=-1
            )
            correction = eyem - torch.real(eyem).to(dtype=q.dtype)
            ret = ret + _TriangularSolve(torch.matmul(q, correction.adjoint()), r)

        return ret

    num_rows, num_cols = q.shape[-2], r.shape[-1]

    if num_rows >= num_cols:
        return _QrGradSquareAndDeepMatrices(q, r, dq, dr)

    y = a[..., :, num_rows:]
    u = r[..., :, :num_rows]
    dv = dr[..., :, num_rows:]
    du = dr[..., :, :num_rows]
    dy = torch.matmul(q, dv)
    dx = _QrGradSquareAndDeepMatrices(q, u, dq + torch.matmul(y, dv.adjoint()), du)
    return torch.cat([dx, dy], dim=-1)


class torchqr(torch.autograd.Function):
    """
    Customized backward of qr for better numerical stability
    """

    @staticmethod
    def forward(ctx, a: Array) -> Any:
        q, r = torch.linalg.qr(a, mode="reduced")
        ctx.save_for_backward(a, q, r)
        return q, r

    @staticmethod
    def backward(ctx, dq: Array, dr: Array) -> Any:
        a, q, r = ctx.saved_tensors
        return torchqr_grad(a, q, r, dq, dr)
